// Copyright 2019-2023 Azul Systems, Inc.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//
// 1. Redistributions of source code must retain the above copyright notice,
// this list of conditions and the following disclaimer.
//
// 2. Redistributions in binary form must reproduce the above copyright notice,
// this list of conditions and the following disclaimer in the documentation
// and/or other materials provided with the distribution.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
// AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
// IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
// ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
// LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
// CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
// SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
// INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
// CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
// ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
// POSSIBILITY OF SUCH DAMAGE.


package jdk.internal.crac.mirror.impl;

import jdk.internal.crac.mirror.Context;
import jdk.internal.crac.mirror.Resource;
import jdk.internal.crac.mirror.RestoreException;
import jdk.internal.crac.mirror.impl.AbstractContext;

import java.util.*;

/**
 * Context performing Checkpoint notification in reverse order of registration.
 * Concurrent registration along notification is silently ignored.
 * @param <R>
 */
public class OrderedContext<R extends Resource> extends AbstractContext<R> {
    private final WeakHashMap<R, Long> resources = new WeakHashMap<>();
    private long order = 0;
    private List<R> restoreSnapshot = null;

    protected List<R> checkpointSnapshot() {
        List<R> snapshot;
        synchronized (this) {
            snapshot = this.resources.entrySet().stream()
                .sorted(Collections.reverseOrder(Map.Entry.comparingByValue()))
                .map(Map.Entry::getKey)
                .toList();
        }
        restoreSnapshot = new ArrayList<>(snapshot);
        Collections.reverse(restoreSnapshot);
        return snapshot;
    }

    // We won't synchronize access to restoreSnapshot because methods
    // beforeCheckpoint and afterRestore should be invoked only
    // by the single thread performing the C/R and other threads should
    // not touch that.
    protected List<R> restoreSnapshot() {
        List<R> snapshot = restoreSnapshot;
        restoreSnapshot = null;
        return snapshot;
    }

    @Override
    public synchronized void register(R resource) {
        resources.put(resource, order++);
    }
}
